import concurrent.futures
from py_actors.actors import Actor
from py_actors.log_config import actor_logger


class MicroKernel(object):

    def __init__(self):
        self.is_running = True
        self.can_submit = True
        self.actor_lookup = dict()
        self.actor_future = dict()
        self.pool = concurrent.futures.ThreadPoolExecutor()
        self.monitor_future = concurrent.futures.Future()

    def submit(self, name, actor_instance):
        if self.can_submit:
            actor_instance.name = name
            self.actor_lookup[name] = actor_instance
            actor_instance.on_init(self.actor_lookup)
            self.actor_future[name] = self.pool.submit(Actor.call, actor_instance)
            return True
        else:
            return False

    def post(self, name, message):
        actor = self.actor_lookup[name]
        actor.post(message)

    def start(self):
        self.monitor_future = self.pool.submit(self.monitor, self)

    @staticmethod
    def monitor(kernel):
        while kernel.is_running or bool(kernel.actor_future):
            for actor_name in list(kernel.actor_future):
                try:
                    is_complete = kernel.actor_future[actor_name].result(timeout=5)
                    actor = kernel.actor_lookup[actor_name]
                    del kernel.actor_future[actor_name]
                    del kernel.actor_lookup[actor_name]
                    if is_complete is True:
                        actor.on_complete()
                except concurrent.futures.TimeoutError:
                    actor_logger.debug(f'Actor(s) {", ".join(list(kernel.actor_future))} has/have not completed yet')
        return True

    def shutdown(self, immediate=True):
        self.can_submit = False
        self.is_running = False
        if immediate:
            actor_logger.debug(f'Shutting down {list(self.actor_lookup)} preemptively')
            for entry in self.actor_lookup.values():
                entry.shutdown()
        self.pool.shutdown()
        return



from threading import Lock


class CallbackFuture:

    def __init__(self):
        self.results = None
        self.mu = Lock()
        self.mu.acquire()

    def callback(self, data):
        try:
            self.results = data
        finally:
            self.mu.release()

    def done(self):
        try:
            self.mu.acquire()
            return self.results
        finally:
            self.mu.release()


class DoneMessage:
    pass


class FlushMessage:
    pass



import abc
import queue
import time
from py_actors.control import DoneMessage, FlushMessage
from py_actors.log_config import actor_logger


class Actor(abc.ABC):

    def __init__(self):
        self.actor_lookup = None
        self.is_running = True
        self.is_complete = False
        self.message_queue = queue.Queue()
        self.name = None

    @classmethod
    def on_receive(cls, message):
        pass

    def get_name(self):
        return self.name

    def on_init(self, lookup):
        self.actor_lookup = lookup

    def on_complete(self):
        actor_logger.info(f'{self.name} has finished')

    def on_shutdown(self):
        actor_logger.info(f'{self.name} was shut down')

    @staticmethod
    def call(actor):
        while actor.is_running:
            if actor.is_complete:
                break
            try:
                message = actor.message_queue.get_nowait()
                actor.on_receive(message)
            except queue.Empty:
                time.sleep(.25)
        actor.on_shutdown()
        return actor.is_complete

    def post(self, message):
        try:
            assert self.is_running is True and self.is_complete is False
            return self.message_queue.put_nowait(message)
        except queue.Full:
            actor_logger.error(f'The queue for {self.name} is full, shutting down')
            self.shutdown()
        except AssertionError:
            actor_logger.error(f'A message was sent to {self.name} after it stopped running or was marked as complete')
            self.shutdown()

    def do_lookup(self, name):
        return self.actor_lookup[name]

    def shutdown(self):
        self.is_running = False


class BatchJoinActor(Actor):

    def __init__(self, key_list):
        super().__init__()
        self.key_list = key_list
        self.num_done = 0
        self.results = []

    def on_receive(self, message):
        if type(message) == DoneMessage:
            self.num_done += 1
            if self.num_done >= len(self.key_list):
                self.is_complete = True
        else:
            self.results.extend(message)


class BatchSplitActor(Actor):

    def __init__(self, key_list, batch_size=256):
        super().__init__()
        self.key_list = key_list
        self.batch_size = batch_size
        self.batch_dict = dict(zip(key_list, [[] for i in range(len(key_list))]))

    def on_receive(self, message):
        if type(message) == DoneMessage:
            for key in self.key_list:
                instance = self.do_lookup(key)
                final_batch = self.batch_dict[key]
                if final_batch:
                    instance.post(final_batch)
                instance.post(DoneMessage())
            self.is_complete = True
        elif type(message) == FlushMessage:
            for key in self.key_list:
                instance = self.do_lookup(key)
                flush_batch = self.batch_dict[key]
                if flush_batch:
                    actor_logger.debug(f'Flushing {flush_batch}')
                    instance.post(flush_batch)
                self.batch_dict[key] = []
        else:
            try:
                message_iterable = iter(message)
                while True:
                    for key in self.key_list:
                        self.batch_dict[key].append(next(message_iterable))
                        if len(self.batch_dict[key]) >= self.batch_size:
                            instance = self.do_lookup(key)
                            instance.post(self.batch_dict[key])
                            self.batch_dict[key] = []
            except StopIteration:
                pass
            except TypeError:
                actor_logger.error(f'{self.name} was given a non-iterable message')
                self.shutdown()


class CountdownActor(Actor):

    def __init__(self, count):
        super().__init__()
        self.count = count

    def on_receive(self, message):
        self.do_work(message)
        self.count -= 1
        if self.count <= 0:
            self.is_complete = True

    @classmethod
    def do_work(cls, message):
        pass


class JoinActor(Actor):

    def __init__(self, key_list):
        super().__init__()
        self.key_list = key_list
        self.num_done = 0
        self.results = []

    def on_receive(self, message):
        if type(message) == DoneMessage:
            self.num_done += 1
            if self.num_done >= len(self.key_list):
                self.is_complete = True
        else:
            self.results.append(message)


class SplitActor(Actor):

    def __init__(self, key_list):
        super().__init__()
        self.key_list = key_list

    def on_receive(self, message):
        if type(message) == DoneMessage:
            for needs_shutdown in self.key_list:
                shut_me_down = self.do_lookup(needs_shutdown)
                shut_me_down.post(DoneMessage())
            self.is_complete = True
        else:
            try:
                message_iterable = iter(message)
                while True:
                    for key in self.key_list:
                        instance = self.do_lookup(key)
                        instance.post(next(message_iterable))
            except StopIteration:
                pass
            except TypeError:
                actor_logger.error(f'{self.name} was given a non-iterable message')
                self.shutdown()






import os
from setuptools import setup


# Utility function to read the README.md file.
def read(filename):
    return open(os.path.join(os.path.dirname(__file__), filename)).read()


setup(
    name="py_actors",
    version="0.0.3",
    author="Kevin Conley",
    author_email="kmanc@comcast.net",
    description="Actor implementation in Python",
    license='MIT',
    keywords="Python, Actor, Concurrency, Parallelism",
    url="https://github.com/kmanc/py_actors",
    packages=['py_actors'],
    long_description=read('README.md'),
    python_requires='>=3.6',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Utilities"
    ],

    setup_requires=[
        'pytest-runner',
    ],
    tests_require=['pytest']
)



import logging

formatter = logging.Formatter('%(asctime)s | %(levelname)s | %(message)s')

actor_logger = logging.getLogger('actor_logger')
actor_file = logging.FileHandler('actor.log')
actor_file.setFormatter(formatter)
actor_logger.addHandler(actor_file)
